/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
// TODO
// * Fermat interface.
// * Adaptive weight bit choice.


#include <cstddef>
#include <utility>
#include <unordered_map>
#include <algorithm> // min, swap
#include <numeric> // accumulate
#include <fstream>
#include <array>
#include <tuple>

#include "integrals.h"

namespace pyred {

/****************
 * Compositions *
 ****************/

int binomial(int n, int k) {
  // Binomial coefficient 'n choose k'.
  if (k < 0 || k > n) return 0;
  int bino = 1;
  for (int i = 0; i < std::min(k, n-k); ++i) {
    bino = (bino * (n-i)) / (i+1);
  }
  return bino;
}

std::vector<std::vector<uint>> compositions(uint n, uint k) {
  // Return a list of all compositions of the integer n into k parts.
  std::vector<std::vector<uint>> compos;
  std::vector<std::vector<uint>> tmp;
  compos.reserve(binomial(n+k-1,k-1));
  tmp.reserve(compos.size());
  compos.emplace_back();
  for (uint i = 1; i < k; ++i) {
    for (const auto &c: compos) {
      auto clen = std::accumulate(c.cbegin(), c.cend(), uint(0));
      for (uint j = 0; j < n+1-clen; ++j) {
        tmp.emplace_back(c);
        tmp.back().emplace_back(j);
      }
    }
    compos.swap(tmp);
    tmp.clear();
  }
  if (k > 0) {
    for (auto &c: compos) {
      c.emplace_back(n-std::accumulate(c.cbegin(), c.cend(), 0));
    }
  }
  return compos;
}

// Cache for get_composition_id().
// Note that pointers/references to values in an unordered_map
// remain valid after insert/erase/rehashing.
std::unordered_map<
  std::pair<uint,uint>,
  std::pair<
    std::unordered_map<std::vector<uint>,uint>,
    std::vector<std::vector<uint>>
  >
> composition_id;


const std::pair<
  std::unordered_map<std::vector<uint>,uint>,
  std::vector<std::vector<uint>>
>& get_composition_id(uint dots, uint lines) {
  // Return the composition weight dictionary 'compo_weights'
  // and the composition from weight list 'compo_from_weight'.
  // compo_weights[compo] = weight and compo_from_weight[weight] = compo
  // for a composition of the integer 'dots' into 'lines' parts.
  static std::mutex mtx;
  {
    std::lock_guard<std::mutex> lock(mtx);
    auto resit = composition_id.find({dots, lines});
    if (resit != composition_id.cend()) {
      return resit->second;
    }
  }
  std::unordered_map<std::vector<uint>,uint> compo_weights;
  std::vector<std::vector<uint>> compo_from_weight;
  uint weight{0};
  auto compos = compositions(dots,lines);
  for (const auto &compo: compos) {
    compo_weights.insert({compo, weight});
    compo_from_weight.push_back(compo);
    ++weight;
  }
  std::lock_guard<std::mutex> lock(mtx);
  // Check if the result has been added to the cache in the meantime.
  // If yes, return the result from the cache to not invalidate
  // references to it.
  auto resit = composition_id.find({dots, lines});
  if (resit != composition_id.cend()) {
    return resit->second;
  }
  // Otherwise add it to the cache and return.
  auto it = composition_id.emplace(
    std::make_pair(dots, lines),
    std::make_pair(compo_weights, compo_from_weight)
  ).first;
  return it->second;
}

/******************
 * Seed generator *
 ******************/

Seeds::Seeds(uint np, uint sect, uint dots, uint sps)
: sector(np, false)
{
  uint lines{0};
  uint nums;
  for (auto pr = sector.begin(); pr != sector.end(); ++pr) {
    if (sect & 1) {
      ++lines;
      *pr = true;
    }
    sect >>= 1;
  }
  nums = np - lines;
  const auto& ppows = get_composition_id(dots,lines).second;
  const auto& spows = get_composition_id(sps,nums).second;
  ppows_begin = ppows.crbegin();
  ppows_end = ppows.crend();
  spows_begin = spows.cbegin();
  spows_end = spows.cend();
}

SeedIterator::SeedIterator()
: sector{}, p_begin{}, p_it{}, p_end{}, s_it{}, s_end{}
{}

SeedIterator::SeedIterator(
  const std::vector<bool>& sector,
  const compos_rit_type& p_begin, const compos_rit_type& p_end,
  const compos_it_type& s_begin, const compos_it_type& s_end)
: sector{sector},
  p_begin{p_begin}, p_it{p_begin}, p_end{p_end}, s_it{s_begin}, s_end{s_end}
{}

SeedIterator& SeedIterator::operator++() {
  if (p_it != p_end) {
    ++p_it;
  }
  if (p_it == p_end) {
    p_it = p_begin;
    ++s_it;
  }
  if (s_it == s_end) {
    // The iterator is depleted.
    sector.clear();
    p_begin = compos_rit_type{};
    p_it = compos_rit_type{};
    p_end = compos_rit_type{};
    s_it = compos_it_type{};
    s_end = compos_it_type{};
  }
  return *this;
}

SeedIterator SeedIterator::operator++(int) {
  auto it = *this;
  ++(*this);
  return it;
}

std::vector<pow_type> SeedIterator::operator*() const {
  std::vector<pow_type> powers;
  powers.reserve(sector.size());
  auto ppows_it = p_it->cbegin();
  auto spows_it = s_it->cbegin();
  for (bool pr: sector) {
    if (pr) {
      powers.push_back(1+(*(ppows_it++)));
    }
    else {
      powers.push_back(-(*(spows_it++)));
    }
  }
  return powers;
}

std::unique_ptr<std::vector<pow_type>> SeedIterator::operator->() const {
  std::unique_ptr<std::vector<pow_type>> tmp{new std::vector<pow_type>};
  *tmp = *(*this);
  return tmp;
}

bool operator==(const SeedIterator& lhs, const SeedIterator& rhs) {
  return (lhs.sector == rhs.sector && lhs.p_begin == rhs.p_begin &&
          lhs.p_it == rhs.p_it && lhs.p_end == rhs.p_end &&
          lhs.s_it == rhs.s_it && lhs.s_end == lhs.s_end);
}

bool operator!=(const SeedIterator& lhs, const SeedIterator& rhs) {
  return !(lhs == rhs);
}

void swap(SeedIterator& lhs, SeedIterator& rhs) {
  std::swap(lhs.sector, rhs.sector);
  std::swap(lhs.p_begin, rhs.p_begin);
  std::swap(lhs.p_it, rhs.p_it);
  std::swap(lhs.p_end, rhs.p_end);
  std::swap(lhs.s_it, rhs.s_it);
  std::swap(lhs.s_end, lhs.s_end);
}

/******************
 * IntegralPyred class *
 ******************/

// static members
uint IntegralPyred::s_np{0};
uint IntegralPyred::s_ordering{0};
std::vector<std::pair<uint,uint>> IntegralPyred::s_sector_weight{};
std::array<uint,6> IntegralPyred::s_weight_bits{{0,0,0,0,0,0}};
std::array<weight_type,6> IntegralPyred::s_weight_proj{{0,0,0,0,0,0}};
std::vector<std::unordered_set<uint>> IntegralPyred::s_trivialsectors{};
uint IntegralPyred::minlines{1};
int IntegralPyred::cache_level{0};
std::unordered_map<IntegralPyred,weight_type> IntegralPyred::i2w_cache{};
std::unordered_map<weight_type,IntegralPyred> IntegralPyred::w2i_cache{};
std::mutex IntegralPyred::i2w_mtx{};
std::mutex IntegralPyred::w2i_mtx{};

IntegralPyred::IntegralPyred(uint topo, const std::vector<pow_type>& pows)
: topology{topo},
  powers{pows}
{
  if (size() != s_np) {
    throw init_error("IntegralPyred(): wrong number of indices");
  }
}

IntegralPyred::IntegralPyred(uint topo, std::vector<pow_type>&& pows)
: topology{topo},
  powers{std::move(pows)}
{
  if (size() != s_np) {
    throw init_error("IntegralPyred(): wrong number of indices");
  }
}

IntegralPyred::IntegralPyred(std::istringstream& ss, uint topo) {
  // Input format: "*[n1,n2,..]*"
  // Skip beginning until '[' is found, stop at ']'.
  topology = topo;
  reserve(s_np);
  pow_type pow;
  char sep;
  do {
    ss >> sep;
  } while (sep != '[');
  sep = ',';
  while (sep == ',') {
    ss >> pow;
    powers.push_back(pow);
    ss >> sep;
  }
  if (sep != ']') {
    throw init_error("IntegralPyred parser: ill-formatted integral");
  }
  if (size() != s_np) {
    throw init_error("IntegralPyred parser: wrong number of indices");
  }
}

IntegralPyred::IntegralPyred(weight_type w) {
  // Convert an integer integral weight to an integral.
  if (!w) {
    throw input_error(
      "IntegralPyred(): cannot construct an integral of zero weight.");
  }
  if (cache_level & 1) {
    std::lock_guard<std::mutex> lock(w2i_mtx);
    auto posit = w2i_cache.find(w);
    if (posit != w2i_cache.cend()) {
      *this = posit->second;
      return;
    }
  }
  auto props = properties(w);
  topology = props.topology;
  powers.reserve(s_np);
  // * dots to the left have lower weight
  // auto ppows_it = get_composition_id(props.dots,props.lines).second.cend()[
  //                   -1-static_cast<int>(props.ppows_weight)].cbegin();
  // * dots to the right have lower weight
  auto ppows_it = get_composition_id(props.dots,props.lines).second[
                    props.ppows_weight].cbegin();
  // * sps to the right have lower weight
  // auto spows_it = get_composition_id(props.sps,props.nums).second[
  //                   props.spows_weight].cbegin();
  // * sps to the left have lower weight
  auto spows_it = get_composition_id(props.sps,props.nums).second.cend()[
                    -1-static_cast<int>(props.spows_weight)].cbegin();
  for (uint dummy = 0; dummy < s_np; ++dummy) {
    if (props.sector & 1) {
      powers.push_back(1+(*(ppows_it++)));
    }
    else {
      powers.push_back(-(*(spows_it++)));
    }
    props.sector >>= 1;
  }
  // Insert the element into the cache(s)
  // if it hasn't been cached in the meantime.
  if (cache_level & 2) {
    std::lock_guard<std::mutex> lock(i2w_mtx);
    i2w_cache.insert({*this, w});
  }
  if (cache_level & 1) {
    std::lock_guard<std::mutex> lock(w2i_mtx);
    w2i_cache.insert({w, *this});
  }
}

IntegralPyredProperties IntegralPyred::properties(weight_type w) {
  // Extract topology, sector, number of lines, numerators,
  // dots, scalar products, dot and sp composition weight.
  // Return them as an IntegralPyredProperties instance.
  // * dot composition has higher weight than sp composition
  // uint spow_w = w & s_weight_proj[5];
  // w >>= s_weight_bits[5];
  // uint ppow_w = w & s_weight_proj[4];
  // w >>= s_weight_bits[4];
  // * sp composition has higher weight than dot composition
  uint ppow_w = w & s_weight_proj[4];
  w >>= s_weight_bits[4];
  uint spow_w = w & s_weight_proj[5];
  w >>= s_weight_bits[5];
  uint dotsp_weight_2 = w & s_weight_proj[3];
  w >>= s_weight_bits[3];
  uint dotsp_weight_1 = w & s_weight_proj[2];
  w >>= s_weight_bits[2];
  uint sect = s_sector_weight[w & s_weight_proj[1]].second;
  w >>= s_weight_bits[1];
  uint sectcpy{sect};
  uint lines{0};
  for (uint dummy = 0; dummy < s_np; ++dummy) {
    if (sectcpy & 1) ++lines;
    sectcpy >>= 1;
  }
  uint nums = s_np - lines;
  uint dots;
  uint sps;
  std::tie(dots, sps) = weight_to_dots(dotsp_weight_1, dotsp_weight_2, lines);
  return IntegralPyredProperties{static_cast<uint>(w), sect, lines, nums,
                            dots, sps, ppow_w, spow_w};
}

std::pair<uint,uint> IntegralPyred::weight_to_dots(
    uint dotsp_1, uint dotsp_2, uint lines) {
  uint dots{0};
  uint sps{0};
  switch (s_ordering) {
    case 1:
      dots = dotsp_1;
      sps = dotsp_2;
      break;
    case 2:
      sps = dotsp_1;
      dots = dotsp_2;
      break;
    case 3:
      dots = dotsp_2;
      sps = dotsp_1 - dotsp_2;
      break;
    case 4:
      dots = dotsp_1 - dotsp_2;
      sps = dotsp_2;
      break;
    case 5:
      dots = dotsp_1 - (lines - minlines);
      sps = dotsp_2;
      break;
  }
  return {dots, sps};
}

void IntegralPyred::setup(uint np, const std::vector<std::unordered_set<uint>>& zs,
                     uint ordering, const std::vector<uint>& weight_bits) {
  static bool setup_done{false};
  static std::mutex setup_mtx;
  std::lock_guard<std::mutex> lock(setup_mtx);
  if (setup_done) {
    throw init_error("IntegralPyred::setup() must only be called once.");
  }
  setup_done = true;
  if (ordering < 1 || ordering > 5) {
    throw init_error("IntegralPyred::setup(): ordering must be one of {1,2,3,4,5}");
  }
  s_np = np;
  s_ordering = ordering;
  uint nontopo_weight = np + weight_bits[0] + weight_bits[1]
                           + weight_bits[2] + weight_bits[3];
  if (nontopo_weight > 64) {
    throw init_error("IntegralPyred weight representation exceeds 64 bits.");
  }
  s_weight_bits[0] = 64 - nontopo_weight; // topology weight: the rest
  s_weight_bits[1] = np; // sector weight
  for (std::size_t n = 2; n < 6; ++n) {
    // weights for dotsp_1, dotsp_2 (meaning depending on ordering),
    // dot distribution, sp distribution
    s_weight_bits[n] = weight_bits[n-2];
  }
  for (std::size_t n = 0; n < 6; ++n) {
    s_weight_proj[n] = (1 << s_weight_bits[n]) - 1;
  }
  // sector weight table
  //   s_sector_weight[sector].first = weight
  //   s_sector_weight[weight].second = sector
  std::vector<std::array<uint,3>> sect_w;
  minlines = s_np;
  for (uint s = 0; s < (1u << s_np); ++s) {
    uint scpy{s};
    uint lines{0};
    uint lines_w{0};
    for (uint l = 0; l < s_np; ++l) {
      if (scpy & 1) {
        ++lines;
      }
      else {
        ++lines_w;
      }
      scpy >>= 1;
      lines_w <<= 1;
    }
    if (s_ordering != 5) {
      sect_w.push_back({{s, lines, lines_w}});
    }
    else {
      sect_w.push_back({{s, 0, s}});
    }
    // minlines: minimal number of lines in any non-trivial sector.
    for (const auto& zs4topo: zs) {
      if (lines < minlines && zs4topo.find(s) != zs4topo.end()) minlines = lines;
    }
  }
  if (zs.empty()) minlines = 0;
  // sort by (number of lines, weight for fixed lines) --> global weight
  std::sort(sect_w.begin(), sect_w.end(),
    [](const std::array<uint,3>& slw1, const std::array<uint,3>& slw2) {
      return slw1[1] == slw2[1] ? slw1[2] < slw2[2] : slw1[1] < slw2[1];
    }
  );
  // assign sector weight and s_sector_weight[weight].second = sector
  for (uint sw = 0; sw < sect_w.size(); ++sw) {
    sect_w[sw][2] = sw;
    s_sector_weight.emplace_back(0, sect_w[sw][0]);
  }
  // sort by sector number
  std::sort(sect_w.begin(), sect_w.end(),
    [](const std::array<uint,3>& slw1, const std::array<uint,3>& slw2) {
      return slw1[0] < slw2[0];
    }
  );
  // assign s_sector_weight[sector].first = weight
  for (uint sw = 0; sw < sect_w.size(); ++sw) {
    s_sector_weight[sw].first = sect_w[sw][2];
  }
  s_trivialsectors = zs;
}

void IntegralPyred::setup(uint np, const std::vector<std::string>& zsfiles,
                     uint ordering, const std::vector<uint>& weight_bits) {
  // Read trivial sectors from the file 'zsfile'
  std::vector<std::unordered_set<uint>> zs;
  for (const auto& zsfile: zsfiles) {
    std::unordered_set<uint> zs4topo;
    if (!zsfile.empty()) {
      std::ifstream instream{zsfile};
      if (!instream.good()) {
        std::ostringstream ss;
        ss << "Failed reading trivial sector file " << zsfile;
        throw init_error(ss.str());
      }
      std::string line;
      std::getline(instream, line);
      std::istringstream ss{line};
      uint sn;
      char sep;
      do {
        ss >> sn;
        zs4topo.insert(sn);
      } while (bool(ss >> sep));
    }
    zs.push_back(zs4topo);
  }
  setup(np, zs, ordering, weight_bits);
}

uint IntegralPyred::np() {
  return s_np;
}

Seeds IntegralPyred::seeds(uint sect, uint dots, uint sps) {
  return Seeds(s_np, sect, dots, sps);
}

weight_type IntegralPyred::to_weight() const {
  // Convert an integral to an integer weight.
  if (cache_level & 2) {
    std::lock_guard<std::mutex> lock(i2w_mtx);
    auto posit = i2w_cache.find(*this);
    if (posit != i2w_cache.cend()) {
      return posit->second;
    }
  }
  weight_type weight{topology};
  uint lines{0};
  uint dots{0};
  uint nums{0};
  uint sps{0};
  uint dotsp_weight_1{0};
  uint dotsp_weight_2{0};
  std::vector<uint> ppows;
  std::vector<uint> spows;
  uint sector{0};
  uint linepos{0};
  for (const auto p: powers) {
    if (p > 0) {
      sector += (1 << linepos);
      ++lines;
      dots += (p-1);
      ppows.push_back(p-1);
    }
    else {
      sps -= p;
      spows.push_back(-p);
    }
    ++linepos;
  }
  if (topology < s_trivialsectors.size()) {
    if (s_trivialsectors[topology].find(sector) !=
        s_trivialsectors[topology].end()) {
      return 0;
    }
  }
  else if (!s_trivialsectors.empty()) {
    std::ostringstream ss;
    ss << "IntegralPyred::to_weight(): zerosectors not defined for topology "
       << topology << ".";
    throw init_error(ss.str());
  }
  weight <<= s_weight_bits[1];
  weight += s_sector_weight[sector].first;
  nums = s_np - lines;
  // dot and sp composition weights
  // * dots to the left have lower weight
  // const auto& cid = get_composition_id(dots,lines);
  // auto ppows_weight = cid.second.size() - cid.first.at(ppows) - 1;
  // * dots to the right have lower weight
  auto ppows_weight = get_composition_id(dots,lines).first.at(ppows);
  // * sps to the right have lower weight
  // auto spows_weight = get_composition_id(sps,nums).first.at(spows);
  // * sps to the left have lower weight
  const auto& cid = get_composition_id(sps,nums);
  auto spows_weight = cid.second.size() - cid.first.at(spows) - 1;
  // (dots, sps) ordering
  switch (s_ordering) {
    case 1:
      dotsp_weight_1 = dots;
      dotsp_weight_2 = sps;
      break;
    case 2:
      dotsp_weight_1 = sps;
      dotsp_weight_2 = dots;
      break;
    case 3:
      dotsp_weight_1 = dots + sps;
      dotsp_weight_2 = dots;
      break;
    case 4:
      dotsp_weight_1 = dots + sps;
      dotsp_weight_2 = sps;
      break;
    case 5:
      dotsp_weight_1 = dots + (lines - minlines);
      dotsp_weight_2 = sps;
      break;
  }
  (weight <<= s_weight_bits[2]) += dotsp_weight_1;
  (weight <<= s_weight_bits[3]) += dotsp_weight_2;
  // * dot composition has higher weight than sp composition
  // (weight <<= s_weight_bits[4]) += ppows_weight;
  // (weight <<= s_weight_bits[5]) += spows_weight;
  // * sp composition has higher weight than dot composition
  (weight <<= s_weight_bits[5]) += spows_weight;
  (weight <<= s_weight_bits[4]) += ppows_weight;
  if (topology > s_weight_proj[0] ||
      dotsp_weight_1 > s_weight_proj[2] || dotsp_weight_2 > s_weight_proj[3] ||
      ppows_weight > s_weight_proj[4] || spows_weight > s_weight_proj[5]) {
    weight_overflow(topology, dotsp_weight_1, dotsp_weight_2,
                    ppows_weight, spows_weight, lines);
  }
  // Insert the element into the cache(s)
  // if it hasn't been cached in the meantime.
  if (cache_level & 2) {
    std::lock_guard<std::mutex> lock(i2w_mtx);
    i2w_cache.insert({*this, weight});
  }
  if (cache_level & 1 && weight) {
    // Do not cache weight=0; constructing IntegralPyred(0) is not allowed.
    std::lock_guard<std::mutex> lock(w2i_mtx);
    w2i_cache.insert({weight, *this});
  }
  return weight;
}

void IntegralPyred::weight_overflow(uint topo, uint dotsp_1, uint dotsp_2,
                              uint ppow_w, uint spow_w, uint lines) {
  // Error handling in case a sub-weight does not fit
  // in the IntegralPyred weight representation, because too few bits
  // are reserved for the particular sub-weight.
  std::ostringstream ss;
  ss << "IntegralPyred weight overflow in";
  uint dots;
  uint sps;
  std::tie(dots, sps) = weight_to_dots(dotsp_1, dotsp_2, lines);
  if (topo > s_weight_proj[0]) {
    ss << " topology (" << topo
       << " does not fit in " << s_weight_bits[0] << " bits);";
  }
  if (dotsp_1 > s_weight_proj[2]) {
    switch (s_ordering) {
      case 1:
        ss << " dots (" << dots
           << " does not fit in " << s_weight_bits[2] << " bits);";
        break;
      case 2:
        ss << " sps (" << sps
           << " does not fit in " << s_weight_bits[2] << " bits);";
        break;
      case 3:
      case 4:
        ss << " dots+sps (" << dots << "+" << sps
           << " does not fit in " << s_weight_bits[2] << " bits);";
        break;
      case 5:
        ss << " dots+lines-minlines ("
           << dots << "+" << lines << "-" << minlines
           << " does not fit in " << s_weight_bits[2] << " bits);";
        break;
    }
  }
  if (dotsp_2 > s_weight_proj[3]) {
    switch (s_ordering) {
      case 1:
      case 4:
      case 5:
        ss << " sps (" << sps
           << " does not fit in " << s_weight_bits[3] << " bits);";
        break;
      case 2:
      case 3:
        ss << " dots (" << dots
           << " does not fit in " << s_weight_bits[4] << " bits);";
        break;
    }
  }
  if (ppow_w > s_weight_proj[4]) {
    ss << " dot distribution (" << ppow_w
       << " does not fit in " << s_weight_bits[4] << " bits);";
  }
  if (spow_w > s_weight_proj[5]) {
    ss << " sp distribution (" << spow_w <<
          " does not fit in " << s_weight_bits[5] << " bits);";
  }
  ss << "\nuse IntegralPyred::setup() to reserve more space where needed.";
  throw init_error(ss.str());
}

IntegralPyred IntegralPyred::zip(const std::vector<pow_type>& merge) const {
  // Return a new IntegralPyred with 'powers' and 'merge' added componentwise.
  std::vector<pow_type> pows;
  pows.reserve(s_np);
  auto p1 = powers.cbegin();
  for (auto p2: merge) {
    pows.push_back(*p1+p2);
    ++p1;
  }
  return IntegralPyred(topology, std::move(pows));
}

void IntegralPyred::use_cache(int level) {
  cache_level = level;
}

void IntegralPyred::clear_cache(int level) {
  if (level & 1) {
    w2i_cache.clear();
  }
  if (level & 2) {
    i2w_cache.clear();
  }
}

bool IntegralPyred::operator<(const IntegralPyred& rhs) const {
  return (to_weight() < rhs.to_weight());
}

bool operator==(const IntegralPyred& lhs, const IntegralPyred& rhs) {
  return (lhs.topology == rhs.topology && lhs.powers == rhs.powers);
}

bool operator!=(const IntegralPyred& lhs, const IntegralPyred& rhs) {
  return (lhs.topology != rhs.topology || lhs.powers != rhs.powers);
}

std::ostream& operator<<(std::ostream& out, const IntegralPyred& igl) {
  out << "T" << igl.topology << igl.powers;
  return out;
}

}
