/* This file is a part of the program Kira.
 * Copyright (C) Johann Usovitsch <jusovitsch@googlemail.com>
 * Philipp Maierhoefer <particle@maierhoefer.net>
 * Peter Uwer <peter.uwer@physik.hu-berlin.de>
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version , or (at
 * your option) any later version as published by the Free Software 
 * Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
*/
#ifndef KIRA_H_
#define KIRA_H_
#include <fstream>
#include <iostream>
#include "ReadYamlFiles.h"
#include "integral.h"
#include "dataBase.h"
#include "connect2kira.h"
#include "interface.h"
#include <tuple>
#include <unordered_map>
#include <gzstream.h>

#define RED 20


typedef std::vector< int > Vint;
typedef Vint::iterator ItVint;
typedef Vint::reverse_iterator RItVint;

typedef std::vector<std::vector<int> > VVint;
typedef VVint::iterator ItVVint;
typedef VVint::reverse_iterator RItVVint;

typedef std::vector<int*> Vintx;
typedef Vintx::iterator ItVintx;

typedef std::vector< Integral * > VG;
typedef VG::iterator ItVG;

typedef std::vector< Equation * > VE;
typedef VE::iterator ItVE;

typedef std::vector< symmetries > SYM;
typedef SYM::iterator ItSYM;

typedef std::vector< std::vector < Integral * > > VVG;
typedef VVG::iterator ItVVG;

typedef std::unordered_map<long unsigned, std::tuple<int,int,int> > INTEGMAP;
typedef INTEGMAP::const_iterator INTEGMAPI;

typedef std::map<long unsigned, std::tuple<Integral ***, int> > FWRED;
typedef FWRED::iterator FWREDI;

typedef std::unordered_map<long unsigned, int > RDY2F;
typedef RDY2F::iterator RDY2FI;

typedef std::map<int, Integral * > map_id_to_key;
typedef map_id_to_key::iterator iterator_id_to_key;

typedef std::list<int> Lint;
typedef std::list<int>::iterator ItLint;

class ConvertResult;
class SeedsKira;
class FUPolynom;
class LeeCriterium;
class info;

struct tupelCharacteristics {
  int sector;
  int topology;
  int flag2;
};

class Kira {
public:
  Kira(std::string, int coreNumber, int algebra_flag, std::string pyred_config);
  friend std::ostream& operator<< (std::ostream& out, const Integral& per);
  friend std::istream& operator>> (std::istream& out, Integral& per);
  friend class ConvertResult;
  friend class Equation;
  friend class SeedsKira;
  friend void initiate_seeds(Kira &kira, std::size_t nOfTopology);
  friend class FUPolynom;
  friend class LeeCriterium;
  // read base data and initiate these base data
  void read_config();
  void read_kinematics();
  void init_kinematics();
  void read_integralfamilies();
  void init_integralfamily();
  void read_reduction_boundaries(Jobs &jobs);
  void collect_reductions(YAML::Node & doc3, std::size_t it);
  void collect_sectors(Jobs &jobs);
  void get_max_bounds();
  void get_topology_relations();
  void execute_jobs();
    
  // create IBP and LI
  void create_IBP();
  void create_IBP_helper(std::vector<GiNaC::possymbol>& var);
  void create_LI();
  void reduce_scal(IBPVG &ibpp);
  
  // get zero sectors
  int find_zero_sectors(int numberPositiveIndices, int sector2Reduce);

  // symmetry relations
  int skip_symmetry(std::string topoName, int j, int it, int itt);
  int symmetry_finder(int it, int itt, std::string topoName,int j, int array[], int array2[], int array3[], int flag);
  int prepare_symmetry();
  void symmetry_relations(std::string topoName);
  int test(GiNaC::lst &matr);
  void write_symmetries(const std::string otputName, SYM symVec[]);
  int read_symmetries(const std::string inputName, SYM symVec[]);
  void print_relations(const std::string name, SYM *&symVec, int itC);

  // setup a system of equations
  void print_equation(Integral*& integral,ogzstream& output, std::pair<int,int>& printControl,const std::string & dest);
  void print_equation2(Integral*& integral,ogzstream& output);
  void write_init_system(std::vector<Equation*>& setUpEq,const std::string & dest, int secNUM,const std::string & add);
  void write_pyred(std::vector<std::vector<std::pair<long unsigned,std::string> > > & eqs,/*
  std::vector<std::vector<int > > & arraySeed,*/
  std::vector<pyred::intid> & independent_eqnums);
  std::vector<std::vector<int > > read_id2int();
  void write_triangular();
  void read_integral(Integral*& integral,igzstream& input, int Eqlength);
  void read_integral2(igzstream& input, int Eqlength);
  void load_pyred(std::vector<std::vector<std::pair<long unsigned,std::string> > > & eqs );
  void load_pyred_on_the_fly(std::vector<pyred::intid> & independent_eqnums,
    std::vector<unsigned long>& mandatory, std::vector<unsigned long>& optional);
  int load_triangular();
  void load_back_substitution();
  void initiate_system_of_equations(int nOfTopology);
  int setup_ibp_li_eq(std::vector<Equation*>& setUpEq, VE &identities, std::vector<long unsigned>& weights, std::string a,int topology);
  int setup_sym_eq(std::vector<Equation*>& setUpEq, std::string a, int topologyNUM, int secNUM);
  int look_up_symmetry(std::vector<Equation*>& setUpEq, IBPIntegral *&r, SYM *&sym, int nOfTopology);
  
  // use pyred
  void run_pyred(std::vector<long unsigned>& mandatory, std::vector<long unsigned>& optional,
		std::vector<info> mandatoryRec, 
		std::vector<info> optionalRec);
  void get_reduced_eqs(pyred::System & sys);
  void record_masters(std::pair<std::vector<pyred::intid>,std::vector<long unsigned> >& subsys);
  
  // run reduction
  std::vector<std::string> collectReductions;
  void initiate_fermat(int kira2math);
  void destroy_fermat();
  void complete_triangular();
  int insert_equation( Integral *&EqPtr, Integral *&insertEqPtr, int id);
  void run_reduction(int opt);
  void sort_equation(Integral *&EqPtr);

  void complete_reduction();
  
  int get_number_of_masterintegrals(unsigned it, std::vector<long unsigned>& finalResults,
				    std::vector<long unsigned>& lriMI,int **IDarray);
  unsigned addup_common_masters(unsigned it, unsigned j, int **IDarray, 
			    Integral newEqPtr[],std::vector<std::list<std::string> > &listPtrToString);
  unsigned addup_final_masters(unsigned it, unsigned j, int **IDarray, 
				Integral newEqPtr[], std::vector<std::list<std::string> >  &listPtrToString);
  void back_subs();
  void clean_back_subs();

  // output mathematica
  void kira_mathematica(std::string,int massRC);
  
private:
  Fermat *fermat;
  long long int core_mask;
  int coreNumber, algebra;
  std::string pyred_config;
  std::string outputDir;
  std::string inputDir;
  
  int OTF, dataBase, dataFile, serialWrite, skipPyred, writeNumericalSystem;
  unsigned maxParallelSectors;
  
  int calculate_coefficient(std::list<std::string>& ptrToString);
  int calculate_coefficient_term(std::list<std::string>& ptrToString);
  std::string jobName,specialName;
  
  // maps
  unsigned numberOfEq, numberOfEqXXX;
  Integral **systemEq;
//   std::vector<Equation*> setUpEq;
//   std::vector<Equation> setUpEq2;
//   Equation** setUpEq;
  
  // zero sectors/symmetries/shifts  
  GiNaC::lst suby;
  std::vector <std::tuple<GiNaC::lst,GiNaC::lst, int> > externalTransf;
  int controlSymmetries;

  // initiate kira, base data
  Integral_F integralfamily;
  std::map<std::string,Integral_F> topology;
  std::string fermatPath;
  int numFlag, biggestBound, realSector;
  std::vector<int> num, numMin, den, denTPlus;
  std::vector<GiNaC::possymbol> externalVar, allVar, invar;
  std::vector<int> invarDim;
  int massSet2OneDim;
  GiNaC::ex dimension, massSet2One;
  GiNaC::lst kinematic, kinematic2, kinematicOld, specialKinematics,
  momentConservation, kinematicShift, kinematicShiftB, kinematicReverse,
  mass2One;

  GiNaC::ex mom_uno;
  GiNaC::possymbol *bS;
  GiNaC::symtab GiNaCSymbols;

  int onlyBacksubstitution, onlyTriangular, onlyPyred, 
      onlyInitiate, startJob;

  
  unsigned reihen;
  Integral ***allEq;
  std::map<long unsigned, std::tuple<Integral ***, int> > forwardRed;
  std::unordered_map<long unsigned,int> rdy2F;
  std::vector<long unsigned> masterVectorSkip;
  std::multimap<std::string,std::vector<int> > selectMastersReduction;
  
  unsigned *length, *rdy2P;
  std::unordered_map<long unsigned,unsigned> occurrence;
  std::vector <long unsigned> reduce_please;
  std::vector <unsigned> reduct2StartHere;
  long unsigned *last_reduce;
  std::unordered_map<long unsigned, unsigned> reverseLastReduce;
  unsigned eqnum;
  
  // seeds:
  SeedsKira *int_seeds;
};

class info{
public:
  int dots;
  int nums;
  int sector;
  int topology;
};

class kiraOutput {
public:
  std::string str[12];
  kiraOutput(int choice){
    switch (choice){
      case 0:
	str[0] = "FORM";
	str[1] = ".inc";
	str[2] = "";
	str[3] = "id ";
	str[4] =  " + ";
	str[5] = "(";
	str[6] =";\n\n";
	str[7] = ") = \n";
	str[8] = ")*(";
	str[9] =")\n";
	str[10] ="0\n";
	str[11] =";";        
	break;
      case 1:
	str[0] = "Mathematica";
	str[1] = ".m";
	str[2] = "{\n";
	str[3] = "";
	str[4] =  " + ";
	str[5] = "[";
	str[6] =",\n";
	str[7] = "] -> \n";
	str[8] = "]*(";
	str[9] =")\n";
	str[10] ="0\n";
	str[11] ="}";        
	break;
    }
  };
};

class ConvertResult{
public:
  ConvertResult(){};
  ConvertResult(Kira& kira, std::string topologyName, int topologyNumber, std::string inputName,const  std::string& inputDir, std::vector<long unsigned> & idOfSeed);
  ~ConvertResult();
  void read_trivial_sector();
  void look_up_seeds();
  int get_sector_number(std::vector<int>& seed);
  void get_integral_id(std::vector<long unsigned> & idOfSeed);
  int skip_integral(std::vector<int>& seed);
  void reconstruct_mass( Kira& kira, std::vector<DBintegral>& integralV );
  void prepare_FORM(std::vector<DBintegral>& integralV);

  int output(Kira& kira, int massReconstruction, std::vector<long unsigned>& idOfSeed, int choice);
  DataBase *database;
  std::string topologyName;
  int topologyNumber;
  std::string inputDir;
  
  std::ofstream Output;
  std::ifstream seedsInput;
  std::ifstream trivialSectorInput;
  std::vector<int> trivSV;
  std::vector<std::vector<int> > arraySeed;
  std::vector<std::vector<int> > zeroIntegrals;
};

#endif // KIRA_H_
